unit TimePanel;

{
  Copyright Notice
  ------------------------------------------------------------------------------
  TTimePanel is  Copyright 1999 Dave Hughes
  EMail : dave@waveform.free-online.co.uk
          hughesd6@cs.man.ac.uk
          dave@hartlib.demon.co.uk
  Web :   http://www.waveform.free-online.co.uk

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  + The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software in a
    product, an acknowledgment in the product documentation would be
    appreciated but is not required.
  + Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.
  + This notice may not be removed or altered from any source distribution.
}

{
  Alteration History (* Information, + Added Functionality, - Bug Fixed)
  ------------------------------------------------------------------------------

  08/10/99 v1.0a
  * First release into the public domain!

  15/10/99 v1.1
  + Thanks ever so much for the alterations to the scrolling routines by
    David Johnson of Futonics (marked DJ in this file), control boxes hitting
    the edge of the control now cause scrolling to occur

  18/10/99 v1.2
  * Tidied up some typos following a stupid botch up with a global
    Find/Replace

  21/10/99 v1.3
  - Corrected problem of hiding the start and finish boxes when duration is
    small by shifting duration box to middle of the "stack" of boxes

  22/10/99 v2.0
  + Added Pause time capability, thanks to Karl-Heinz Stimmel for the idea
  + Added Align property
  + Added standard TPanel events
  + Added Simple time capability (one time only - no duration / pause etc.),
    thanks again to David Johnson for the idea (and a hint for implementation!)
  * Removed TimeFormat, and replaced with extra properties StartFormat and
    FinishFormat
  * Restructured property default code to allow for easier alteration
  * Optimised the drag code
  * Removed unnecessarily declared type
  * Restructured class hierarchy with TCustomTimePanel
  - Fixed problems with setting StartTime and FinishTime when parameter has
    date information included
  - Removed useless TabStop and TabOrder properties (component cannot receive
    keyboard focus, so these are pointless)
  - Removed useless Color property (cannot currently change background colour)

  26/10/99 v2.1
  + Added PauseTime property (pause duration in a TDateTime), thanks to
    Karl-Heinz Stimmel for the ideas on the new properties
  + Added DurationTime property (total duration in a TDateTime)
  + Added TimeShift property to control how many minutes the values change by
    when their respective boxes are dragged
  + Added some more properties that I'd missed (like visible ... doh!)
  * Altered behaviour of OnTimeChange to include alteration of TDateTime
    properties

  29/11/99 v3.0
  + Restructured control completely to aid editing of properties at design time
    (the wealth of properties was becoming a mite overwhelming plus it makes
    the TCustomTimePanel code a bit neater)
  + Added Alignment property which controls alignment of the time boxes within
    the control
  + Added Connect property to change the properties of the connecting
    line between the start and finish boxes
  + Added ShowFocus property to draw dotted line round control when it has
    keyboard focus (not that keyboard focus will do much for you at the moment)
  + Added some stuff in the Paint routine to make the control appear different
    when it's disabled
  + Added TDBTimePanel (this was another reason for the restructure)
  * Apologies for the inconvenience caused by the restructure. I realise this
    means people will have to re-create the control in most cases - but
    trust me! It's for the best!
}

interface

{--- Set up compiler defines for Delphi versions}
{$DEFINE D2}
{$IFNDEF VER90}
  {$DEFINE D3}
  {$IFNDEF VER100}
    {$DEFINE D4}
  {$ENDIF}
{$ENDIF}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, DB, DBCtrls {$IFNDEF D3}, DBTables {$ENDIF};

type

  { Custom types used by TTimePanel }

  TTPDays = 1..2;
  TTPHours = 0..23;
  TTPDoubleHours = 0..47;
  TTPMins = 0..59;
  TTPTimeShift = 2..15;
  TTPArrowDirection = (adUp, adDown);
  TTPScrollDirection = (sdUp, sdDown, sdNotScrolling);
  TTPControlKind = (ckStart, ckFinish, ckPause, ckDuration, ckSimple);

  { TCustomTimePanel forward declaration }

  TCustomTimePanel = class;

  { TCustomTimeBox }

  TCustomTimeBox = class(TPersistent)
  private
    { Private declarations }
    FPanel: TCustomTimePanel;
    FKind: TTPControlKind;
    FColor: TColor;
    FFormat: String;
    FHours: TTPHours;
    FMins: TTPMins;
    FDay: TTPDays;
    FRect: TRect;
    FShown: Boolean;
    FVisible: Boolean;
    function GetDay: TTPDays;
    function GetHours: TTPHours;
    function GetMins: TTPMins;
    function GetTime: TDateTime;
    procedure SetColor(Value: TColor);
    procedure SetFormat(Value: String);
    procedure SetDay(Value: TTPDays);
    procedure SetHours(Value: TTPHours);
    procedure SetMins(Value: TTPMins);
    procedure SetTime(Value: TDateTime);
    procedure SetVisible(Value: Boolean);
  protected
    { Protected declarations }
    constructor Create(AOwner: TCustomTimePanel);
    property ControlKind: TTPControlKind read FKind;
    property Rect: TRect read FRect;
    property Shown: Boolean read FShown;
    property Color: TColor read FColor write SetColor;
    property DisplayFormat: String read FFormat write SetFormat;
    property Day: TTPDays read GetDay write SetDay;
    property Hours: TTPHours read GetHours write SetHours;
    property Mins: TTPMins read GetMins write SetMins;
    property Time: TDateTime read GetTime write SetTime;
    property Visible: Boolean read FVisible write SetVisible;
  public
    { Public declarations }
    procedure IncTime(Hours: TTPHours; Mins: TTPMins);
    procedure DecTime(Hours: TTPHours; Mins: TTPMins);
  end;

  { TTimeBox }

  TTimeBox = class(TCustomTimeBox)
  public
    { Public declarations }
    property ControlKind;
    property Rect;
    property Time;
    property Shown;
  published
    { Published declarations }
    property Color;
    property DisplayFormat;
    property Hours;
    property Mins;
  end;

  { TPDTimeBox }

  TPDTimeBox = class(TTimeBox)
  published
    { Published declarations }
    property Visible;
  end;

  { TFTimeBox }

  TFTimeBox = class(TTimeBox)
  published
    { Published declarations }
    property Day;
  end;

  { TTimeBoxConnect }

  TTimeBoxConnect = class(TPersistent)
  private
    { Private declarations }
    FPanel: TCustomTimePanel;
    FWidth: Integer;
    FTopColor: TColor;
    FBottomColor: TColor;
    FVisible: Boolean;
    procedure SetWidth(Value: Integer);
    procedure SetTopColor(Value: TColor);
    procedure SetBottomColor(Value: TColor);
    procedure SetVisible(Value: Boolean);
  public
    { Public declarations }
    constructor Create(AOwner: TCustomTimePanel);
  published
    { Published declarations }
    property Width: Integer read FWidth write SetWidth;
    property TopColor: TColor read FTopColor write SetTopColor;
    property BottomColor: TColor read FBottomColor write SetBottomColor;
    property Visible: Boolean read FVisible write SetVisible;
  end;

  { TTimeBoxProxy }

  TTimeBoxProxy = class(TGraphicControl)
  protected
    constructor Create(AOwner: TComponent); override;
    property OnMouseMove;
  end;

  { TCustomTimePanel }

  TCustomTimePanel = class(TCustomPanel)
  private
    { Private declarations }
    FBottomArrowRect: TRect;
    FBottomDay: TTPDays;
    FBottomHour: TTPHours;
    FConnect: TTimeBoxConnect;
    FDuration: TPDTimeBox;
    FDurationControl: TTimeBoxProxy;
    FFinish: TFTimeBox;
    FFinishControl: TTimeBoxProxy;
    FMarginColor: TColor;
    FPause: TPDTimeBox;
    FPauseControl: TTimeBoxProxy;
    FTextHeight: Integer;
    FTimeColor: TColor;
    FTopArrowRect: TRect;
    FTopDay: TTPDays;
    FTopHour: TTPHours;
    FTopMin: TTPMins;
    FScrollAmount: Single;
    FScrolling: TTPScrollDirection;
    FSelColor: TColor;
    FShowDays: TTPDays;
    FShowFocus: Boolean;
    FShowMargin: Boolean;
    FSimple: TTimeBox;
    FSimpleControl: TTimeBoxProxy;
    FSimplePanel: Boolean;
    FStart: TTimeBox;
    FStartControl: TTimeBoxProxy;
    FTimeShift: TTPTimeShift;
    FWorkFinish: TTPHours;
    FWorkStart: TTPHours;
    FWorkTimeColor: TColor;
    FOnChange: TNotifyEvent;
    function IsInRect(ARect: TRect; X, Y: Integer): Boolean;
    procedure CalcRange;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure DrawArrow(C: TCanvas; Dir: TTPArrowDirection; AColor: TColor);
    procedure DrawConnect(C: TCanvas);
    procedure DrawFrame(C: TCanvas);
    function DrawTimeBox(C: TCanvas; Day: TTPDays; Hours: TTPHours; Mins: TTPMins; AColor: TColor; const BoxText: String): TRect;
    procedure DrawTimeBoxes(C: TCanvas);
    procedure DrawTimeNumbers(C: TCanvas);
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure SetMarginColor(Value: TColor);
    procedure SetSelColor(Value: TColor);
    procedure SetSimplePanel(Value: Boolean);
    procedure SetShowDays(Value: TTPDays);
    procedure SetShowFocus(Value: Boolean);
    procedure SetShowMargin(Value: Boolean);
    procedure SetTimeColor(Value: TColor);
    procedure SetTopDay(Value: TTPDays);
    procedure SetTopHour(Value: TTPHours);
    procedure SetTopMin(Value: TTPMins);
    procedure SetWorkFinish(Value: TTPHours);
    procedure SetWorkStart(Value: TTPHours);
    procedure SetWorkTimeColor(Value: TColor);
    procedure Scroll(Dir: TTPScrollDirection; Hours: TTPHours; Mins: TTPMins);
    procedure WMKillFocus(var Message: TMessage); message WM_KILLFOCUS;
    procedure WMSetFocus(var Message: TMessage); message WM_SETFOCUS;
  protected
    { Protected declarations }
    function CanEdit(ControlKind: TTPControlKind): Boolean; virtual;
    procedure AssignMins(var StMins, FinMins, PaMins, SiMins: Integer);
    procedure CalcMins(var StMins, FinMins, PaMins, DurMins, SiMins: Integer);
    procedure CreateTimeBoxes; dynamic;
    procedure DoChange; virtual;
    procedure DurationMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer); virtual;
    procedure FinishMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer); virtual;
    procedure FixMins(var StMins, FinMins, PaMins, SiMins: Integer);
    procedure Paint; override;
    procedure PauseMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer); virtual;
    procedure SimpleMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer); virtual;
    procedure StartMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer); virtual;
    property Connect: TTimeBoxConnect read FConnect write FConnect;
    property Duration: TPDTimeBox read FDuration write FDuration;
    property Finish: TFTimeBox read FFinish write FFinish;
    property MarginColor: TColor read FMarginColor write SetMarginColor;
    property Pause: TPDTimeBox read FPause write FPause;
    property SelColor: TColor read FSelColor write SetSelColor;
    property ShowDays: TTPDays read FShowDays write SetShowDays;
    property ShowFocus: Boolean read FShowFocus write SetShowFocus;
    property ShowMargin: Boolean read FShowMargin write SetShowMargin;
    property Simple: TTimeBox read FSimple write FSimple;
    property SimplePanel: Boolean read FSimplePanel write SetSimplePanel;
    property Start: TTimeBox read FStart write FStart;
    property TimeColor: TColor read FTimeColor write SetTimeColor;
    property TimeShift: TTPTimeShift read FTimeShift write FTimeShift;
    property TopDay: TTPDays read FTopDay write SetTopDay;
    property TopHour: TTPHours read FTopHour write SetTopHour;
    property TopMin: TTPMins read FTopMin write SetTopMin;
    property WorkFinish: TTPHours read FWorkFinish write SetWorkFinish;
    property WorkStart: TTPHours read FWorkStart write SetWorkStart;
    property WorkTimeColor: TColor read FWorkTimeColor write SetWorkTimeColor;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

  { TTimePanel }

  TTimePanel = class(TCustomTimePanel)
  published
    { Published declarations }
    property Align;
    property Alignment;
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BorderStyle;
    property BorderWidth;
    property Color;
    property Connect;
    property Cursor;
    property Duration;
    property Enabled;
    property Finish;
    property Font;
    property Height;
    property MarginColor;
    property ParentFont;
    property ParentShowHint;
    property Pause;
    property PopupMenu;
    property SelColor;
    property ShowDays;
    property ShowFocus;
    property ShowHint;
    property ShowMargin;
    property Simple;
    property SimplePanel;
    property Start;
    property TabOrder;
    property TabStop;
    property TimeColor;
    property TimeShift;
    property TopDay;
    property TopHour;
    property TopMin;
    property Width;
    property Visible;
    property WorkFinish;
    property WorkStart;
    property WorkTimeColor;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDrag;
    property OnChange;
  end;

  { TDBTimePanel }

  TDBTimePanel = class(TTimePanel)
  private
    { Private declarations }
    FDurationData: TFieldDataLink;
    FFinishData: TFieldDataLink;
    FPauseData: TFieldDataLink;
    FSimpleData: TFieldDataLink;
    FStartData: TFieldDataLink;
    FWriting: Boolean;
    function GetDurationDataField: String;
    function GetDurationField: TField;
    function GetFinishDataField: String;
    function GetFinishField: TField;
    function GetDataSource: TDataSource;
    function GetPauseDataField: String;
    function GetPauseField: TField;
    function GetReadOnly: Boolean;
    function GetSimpleDataField: String;
    function GetSimpleField: TField;
    function GetStartDataField: String;
    function GetStartField: TField;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure CMGetDataLink(var Message: TMessage); message CM_GETDATALINK;
    procedure DataChange(Sender: TObject);
    procedure SetDurationDataField(Value: String);
    procedure SetFinishDataField(Value: String);
    procedure SetPauseDataField(Value: String);
    procedure SetSimpleDataField(Value: String);
    procedure SetStartDataField(Value: String);
    procedure SetDataSource(Value: TDataSource);
    procedure SetReadOnly(Value: Boolean);
    procedure UpdateData(Sender: TObject);
  protected
    { Protected declarations }
    function CanEdit(ControlKind: TTPControlKind): Boolean; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property DurationField: TField read GetDurationField;
    property FinishField: TField read GetFinishField;
    property PauseField: TField read GetPauseField;
    property SimpleField: TField read GetSimpleField;
    property StartField: TField read GetStartField;
  published
    { Published declarations }
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property DurationDataField: String read GetDurationDataField write SetDurationDataField;
    property FinishDataField: String read GetFinishDataField write SetFinishDataField;
    property PauseDataField: String read GetPauseDataField write SetPauseDataField;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly;
    property SimpleDataField: String read GetSimpleDataField write SetSimpleDataField;
    property StartDataField: String read GetStartDataField write SetStartDataField;
  end;

procedure Register;

implementation

uses DsgnIntf, TypInfo;

{ ************************************************************* TTimeBoxProxy }

constructor TTimeBoxProxy.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csOpaque];
  Visible := True;
end; {Create}

{ ************************************************************ TTimeBoxConnect }

constructor TTimeBoxConnect.Create(AOwner: TCustomTimePanel);
begin
  inherited Create;
  FPanel := (AOwner as TCustomTimePanel);
end; {Create}

procedure TTimeBoxConnect.SetTopColor(Value: TColor);
begin
  if Value <> FTopColor then begin
    FTopColor := Value;
    FPanel.Invalidate;
  end;
end; {SetTopColor}

procedure TTimeBoxConnect.SetBottomColor(Value: TColor);
begin
  if Value <> FBottomColor then begin
    FBottomColor := Value;
    FPanel.Invalidate;
  end;
end; {SetBottomColor}

procedure TTimeBoxConnect.SetVisible(Value: Boolean);
begin
  if Value <> FVisible then begin
    FVisible := Value;
    FPanel.Invalidate;
  end;
end; {SetVisible}

procedure TTimeBoxConnect.SetWidth(Value: Integer);
begin
  if Value <> FWidth then begin
    FWidth := Value;
    FPanel.Invalidate;
  end;
end; {SetWidth}

{ ************************************************************* TCustomTimeBox }

constructor TCustomTimeBox.Create(AOwner: TCustomTimePanel);
begin
  inherited Create;
  FPanel := AOwner;
end; {Create}

{ TCustomTimeBox ----------------------------------------- Property Handlers --}

// GetDay returns one unless the control kind is Finish
function TCustomTimeBox.GetDay: TTPDays;
begin
  if FKind = ckFinish then Result := FDay else Result := 1;
end; {GetDay}

// GetHours returns the FHours value, unless the control kind is Duration in
// which case the value is calculated
function TCustomTimeBox.GetHours: TTPHours;
var
  StMins, FinMins, PaMins, DurMins, SiMins: Integer;
begin
  if FKind in [ckStart, ckPause, ckFinish, ckSimple] then
    Result := FHours
  else begin
    FPanel.CalcMins(StMins, FinMins, PaMins, DurMins, SiMins);
    Result := DurMins div 60;
  end;
end; {GetHours}

// GetMins returns the FMins value, unless the control kind is Duration in
// which case the value is calculated
function TCustomTimeBox.GetMins: TTPMins;
var
  StMins, FinMins, PaMins, DurMins, SiMins: Integer;
begin
  if FKind in [ckStart, ckPause, ckFinish, ckSimple] then
    Result := FMins
  else begin
    FPanel.CalcMins(StMins, FinMins, PaMins, DurMins, SiMins);
    Result := DurMins mod 60;
  end;
end; {GetMins}

// GetTime returns the EncodeTime result of FHours and FMins unless the control
// kind is Duration in which case the value is the EncodeTime result of the
// calculated hours and minutes
function TCustomTimeBox.GetTime: TDateTime;
var
  StMins, FinMins, PaMins, DurMins, SiMins: Integer;
begin
  if FKind in [ckStart, ckPause, ckFinish, ckSimple] then
    Result := EncodeTime(FHours, FMins, 0, 0)
  else begin
    FPanel.CalcMins(StMins, FinMins, PaMins, DurMins, SiMins);
    Result := EncodeTime(DurMins div 60, DurMins mod 60, 0, 0);
  end;
end; {GetTime}

// SetColor simply alters the TimeBox's colour and invalidates the parent
procedure TCustomTimeBox.SetColor(Value: TColor);
begin
  if FColor <> Value then begin
    FColor := Value;
    FPanel.Invalidate;
  end;
end; {SetColor}

// SetFormat simply alters the TimeBox's DisplayFormat and invalidates the
// parent
procedure TCustomTimeBox.SetFormat(Value: String);
begin
  if FFormat <> Value then begin
    FFormat := Value;
    FPanel.Invalidate;
  end;
end; {SetFormat}

// SetVisible set's the control's Visible value, and redraws the parent
procedure TCustomTimeBox.SetVisible(Value: Boolean);
begin
  if FVisible <> Value then begin
    FVisible := Value;
    FPanel.Invalidate;
  end;
end; {SetVisible}

// SetDay only does anything with the Finish control. With this control it
// set's the day value if ShowDays permits
procedure TCustomTimeBox.SetDay(Value: TTPDays);
var
  StMins, FinMins, PaMins, DurMins, SiMins: Integer;
begin
  if (FKind = ckFinish) and (FDay <> Value) and (Value <= FPanel.ShowDays) and
    FPanel.CanEdit(ckFinish) then
    with FPanel do begin
      FDay := Value;
      CalcMins(StMins, FinMins, PaMins, DurMins, SiMins);
      FixMins(StMins, FinMins, PaMins, SiMins);
      AssignMins(StMins, FinMins, PaMins, SiMins);
    end; {With}
end; {SetDay}

// SetHours simply set's the control's Hours value, limits the ranges and
// redraws the parent
procedure TCustomTimeBox.SetHours(Value: TTPHours);
var
  StMins, FinMins, PaMins, DurMins, SiMins: Integer;
begin
  if (GetHours <> Value) and (FPanel.CanEdit(FKind)) then
    with FPanel do begin
      if FKind <> ckDuration then begin
        FHours := Value;
        CalcMins(StMins, FinMins, PaMins, DurMins, SiMins);
      end else begin
        CalcMins(StMins, FinMins, PaMins, DurMins, SiMins);
        FinMins := StMins + PaMins + (Value * 60) + (DurMins mod 60);
      end; {If}
      FixMins(StMins, FinMins, PaMins, SiMins);
      AssignMins(StMins, FinMins, PaMins, SiMins);
    end; {With}
end; {SetHours}

// SetMins simply set's the control's Mins value, limits the ranges and
// redraws the parent
procedure TCustomTimeBox.SetMins(Value: TTPMins);
var
  StMins, FinMins, PaMins, DurMins, SiMins: Integer;
begin
  if (GetMins <> Value) and (FPanel.CanEdit(FKind)) then
    with FPanel do begin
      if FKind <> ckDuration then begin
        FMins := Value;
        CalcMins(StMins, FinMins, PaMins, DurMins, SiMins);
      end else begin
        CalcMins(StMins, FinMins, PaMins, DurMins, SiMins);
        FinMins := StMins + PaMins + (DurMins div 60) + Value;
      end; {If}
      FixMins(StMins, FinMins, PaMins, SiMins);
      AssignMins(StMins, FinMins, PaMins, SiMins);
    end; {With}
end; {SetMins}

// SetTime set's the control's Hours and Mins values, limits the ranges
// and redraws the parent
procedure TCustomTimeBox.SetTime(Value: TDateTime);
var
  H, M, S, MS: Word;
  StMins, FinMins, PaMins, DurMins, SiMins: Integer;
begin
  if (GetTime <> Frac(Value)) and FPanel.CanEdit(FKind) then
    with FPanel do begin
      DecodeTime(Value, H, M, S, MS);
      if FKind <> ckDuration then begin
        FHours := H;
        FMins := M;
        CalcMins(StMins, FinMins, PaMins, DurMins, SiMins);
      end else begin
        DecodeTime(Value, H, M, S, MS);
        CalcMins(StMins, FinMins, PaMins, DurMins, SiMins);
        FinMins := StMins + PaMins + (H * 60) + M;
      end; {If}
      FixMins(StMins, FinMins, PaMins, SiMins);
      AssignMins(StMins, FinMins, PaMins, SiMins);
    end; {With}
end; {SetTime}

{---
 Note : In the next two routines, Inc or Dec on ckDuration is a little
 misleading semantically. Basically, you should think of Inc and Dec as acting
 on the time boxes in the control. So Inc and Dec on ckDuration, actually
 Incs and Decs the Start and Finish times, sliding the whole range along.
---}

// IncTime increases the control's time values by the specified numbers of
// hours and minutes, then redraws the parent after re-limiting the ranges
procedure TCustomTimeBox.IncTime(Hours: TTPHours; Mins: TTPMins);
var
  StMins, FinMins, PaMins, DurMins, SiMins: Integer;
begin
  with FPanel do begin
    CalcMins(StMins, FinMins, PaMins, DurMins, SiMins);
    if FKind <> ckDuration then begin
      if FPanel.CanEdit(FKind) then begin
        case FKind of
          ckStart:    Inc(StMins, (Hours * 60) + Mins);
          ckFinish:   Inc(FinMins, (Hours * 60) + Mins);
          ckPause:    Inc(PaMins, (Hours * 60) + Mins);
          ckSimple:   Inc(SiMins, (Hours * 60) + Mins);
        end; {Case}
        FixMins(StMins, FinMins, PaMins, SiMins);
        AssignMins(StMins, FinMins, PaMins, SiMins);
      end; {If}
    end else begin
      if FPanel.CanEdit(ckStart) and FPanel.CanEdit(ckFinish) then begin
        Inc(StMins, (Hours * 60) + Mins);
        Inc(FinMins, (Hours * 60) + Mins);
        FixMins(StMins, FinMins, PaMins, SiMins);
        AssignMins(StMins, FinMins, PaMins, SiMins);
      end; {If}
    end; {If}
  end; {With}
end; {IncTime}

// DecTime decreases the control's time values by the specified numbers of
// hours and minutes, then redraws the parent after re-limiting the ranges
procedure TCustomTimeBox.DecTime(Hours: TTPHours; Mins: TTPMins);
var
  StMins, FinMins, PaMins, DurMins, SiMins: Integer;
begin
  with FPanel do begin
    CalcMins(StMins, FinMins, PaMins, DurMins, SiMins);
    if FKind <> ckDuration then begin
      if FPanel.CanEdit(FKind) then begin
        case FKind of
          ckStart:    Dec(StMins, (Hours * 60) + Mins);
          ckFinish:   Dec(FinMins, (Hours * 60) + Mins);
          ckPause:    Dec(PaMins, (Hours * 60) + Mins);
          ckSimple:   Dec(SiMins, (Hours * 60) + Mins);
        end; {Case}
        FixMins(StMins, FinMins, PaMins, SiMins);
        AssignMins(StMins, FinMins, PaMins, SiMins);
      end; {If}
    end else begin
      if FPanel.CanEdit(ckStart) and FPanel.CanEdit(ckFinish) then begin
        Dec(StMins, (Hours * 60) + Mins);
        Dec(FinMins, (Hours * 60) + Mins);
        FixMins(StMins, FinMins, PaMins, SiMins);
        AssignMins(StMins, FinMins, PaMins, SiMins);
      end; {If}
    end; {If}
  end; {With}
end; {DecTime}

{ *********************************************************** TCustomTimePanel }

constructor TCustomTimePanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csAcceptsControls, csSetCaption] +
    [csReplicatable];
  // Set defaults
  FTopDay := 1;
  FTopHour := 8;
  FTopMin := 0;
  FMarginColor := clBtnFace;
  FScrolling := sdNotScrolling;
  FSelColor := clTeal;
  FShowDays := 2;
  FShowFocus := False;
  FShowMargin := True;
  FSimplePanel := False;
  FTimeColor := clGray;
  FTimeShift := 5;
  FWorkTimeColor := clBlack;
  FWorkStart := 9;
  FWorkFinish := 17;
  // Create connector
  FConnect := TTimeBoxConnect.Create(Self);
  with FConnect do begin
    FTopColor := clGray;
    FBottomColor := clBlack;
    FVisible := True;
    FWidth := 4;
  end;
  // Create start, finish etc. boxes
  CreateTimeBoxes;
  // Set inherited properties defaults
  Caption := '';
  Width := 128;
  Height := 200;
  BevelInner := bvNone;
  BevelOuter := bvRaised;
  BevelWidth := 1;
  BorderStyle := bsNone;
  BorderWidth := 0;
  Color := clBtnFace;
  TabStop := True;
  Font.Name := 'MS Sans Serif';
  Font.Size := 8;
  Font.Style := [];
end; {Create}

procedure TCustomTimePanel.CreateTimeBoxes;
begin
  FSimpleControl := TTimeBoxProxy.Create(Self);
  with FSimpleControl do begin
    Parent := Self;
    OnMouseMove := SimpleMouseMove;
  end;
  FSimple := TTimeBox.Create(Self);
  with FSimple do begin
    FHours := 9;
    FMins := 30;
    FColor := clAqua;
    FFormat := 'HH:mm';
    FDay := 1;
    FKind := ckSimple;
  end;
  FDurationControl := TTimeBoxProxy.Create(Self);
  with FDurationControl do begin
    Parent := Self;
    OnMouseMove := DurationMouseMove;
  end;
  FDuration := TPDTimeBox.Create(Self);
  with FDuration do begin
    FColor := clYellow;
    FFormat := 'D %2dh %.2dm';
    FKind := ckDuration;
    FVisible := True;
  end;
  FPauseControl := TTimeBoxProxy.Create(Self);
  with FPauseControl do begin
    Parent := Self;
    OnMouseMove := PauseMouseMove;
  end;
  FPause := TPDTimeBox.Create(Self);
  with FPause do begin
    FColor := clAqua;
    FFormat := 'P %2dh %.2dm';
    FHours := 0;
    FMins := 30;
    FDay := 1;
    FKind := ckPause;
    FVisible := True;
  end;
  FStartControl := TTimeBoxProxy.Create(Self);
  with FStartControl do begin
    Parent := Self;
    OnMouseMove := StartMouseMove;
  end;
  FStart := TTimeBox.Create(Self);
  with FStart do begin
    FColor := clLime;
    FFormat := 'HH:mm';
    FHours := 9;
    FMins := 0;
    FDay := 1;
    FKind := ckStart;
  end;
  FFinishControl := TTimeBoxProxy.Create(Self);
  with FFinishControl do begin
    Parent := Self;
    OnMouseMove := FinishMouseMove;
  end;
  FFinish := TFTimeBox.Create(Self);
  with FFinish do begin
    FColor := clRed;
    FFormat := 'HH:mm';
    FHours := 11;
    FMins := 30;
    FDay := 1;
    FKind := ckFinish;
  end;
end;

destructor TCustomTimePanel.Destroy;
begin
  FStart.Free;
  FFinish.Free;
  FDuration.Free;
  FPause.Free;
  FSimple.Free;
  FConnect.Free;
  inherited Destroy;
end;

{ TCustomTimePanel ----------------------------------- Time Control Routines --}

// CanEdit does nothing in TCustomTimePanel but can be overriden in derivatives
// to prevent editing under certain situations (e.g. in TDBTimePanel when
// ReadOnly is true. Just make sure you DON'T call inherited!
function TCustomTimePanel.CanEdit(ControlKind: TTPControlKind): Boolean;
begin
  Result := True;
end; {CanEdit}

// CalcMins simply calculates the minutes for Start, Finish, Pause and Duration
// from the hours & minutes values
procedure TCustomTimePanel.CalcMins(var StMins, FinMins, PaMins,
  DurMins, SiMins: Integer);
begin
  with Start do StMins := (Hours * 60) + Mins;
  with Pause do PaMins := (Hours * 60) + Mins;
  with Finish do FinMins := ((Day - 1) * (24 * 60)) + (Hours * 60) + Mins;
  with Simple do SiMins := (Hours * 60) + Mins;
  DurMins := (FinMins - StMins) - PaMins;
end; {CalcMins}

// FixMins corrects any boundary violations in the Start, Finish, and Pause
// minutes parameters. Duration is not included as this is never actually
// stored, just calculated
procedure TCustomTimePanel.FixMins(var StMins, FinMins, PaMins,
  SiMins: Integer);
begin
  if not FSimplePanel then begin
    // Check FinMins isn't too big
    case ShowDays of
      1: if (FinMins > ((24 * 60) - 1)) then FinMins := (24 * 60) - 1;
      2: if (FinMins > ((2 * 24 * 60) - 1)) then FinMins := (2 * 24 * 60) - 1;
    end;
    // Check all values aren't too small
    if StMins < 0 then StMins := 0;
    if FinMins < 0 then FinMins := 0;
    if PaMins < 0 then PaMins := 0;

    // Check St->Fin isn't > 24 hours
    if (FinMins - StMins) > ((24 * 60) - 1) then
      FinMins := StMins + ((24 * 60) - 1);

    // Check St<Fin and pause is small enough for duration
    if FinMins < StMins then begin
      FinMins := StMins;
      PaMins := 0;
    end else if FinMins < (StMins + PaMins) then
      PaMins := FinMins - StMins;
  end else begin
    // Check range of Si
    if SiMins < 0 then SiMins := 0
    else if SiMins > ((24 * 60) - 1) then SiMins := (24 * 60) - 1;
  end;
end; {FixMins}

// AssignMins assigns the values passed in the paramters to the Start, Finish
// and Pause controls in the TimePanel
procedure TCustomTimePanel.AssignMins(var StMins, FinMins, PaMins,
  SiMins: Integer);
begin
  if not FSimplePanel then begin
    with Start do begin
      FHours := StMins div 60;
      FMins := StMins mod 60;
    end;
    with Finish do begin
      FDay := (FinMins div (24 * 60)) + 1;
      FHours := (FinMins div 60) mod 24;
      FMins := FinMins mod 60;
    end;
    with Pause do begin
      FHours := PaMins div 60;
      FMins := PaMins mod 60;
    end;
  end else begin
    with Simple do begin
      FHours := SiMins div 60;
      FMins := SiMins mod 60;
    end;
  end;
  Invalidate;
  DoChange;
end; {AssignMins}

{ TCustomTimePanel ------------------------------------------ Paint Handlers --}

const
  ArrowHeight = 12;
  BoxMargin = 18;
  SingleBorderMargin = 5;
  TimeNumMargin = 4;

procedure TCustomTimePanel.DrawArrow(C: TCanvas; Dir: TTPArrowDirection;
  AColor: TColor);
var
  Arrow: Array [1..3] of TPoint;
  Margin: Integer;
begin
  with C do begin
    Brush.Color := AColor;
    Brush.Style := bsSolid;
    Pen.Color := AColor;
    Pen.Style := psSolid;
  end;
  if Dir = adUp then begin
    Arrow[1] := Point(2, 8);
    Arrow[2] := Point(8, 2);
    Arrow[3] := Point(14, 8);
    with FTopArrowRect do begin
      Top := 2;
      Left := 2;
      Right := 14;
      Bottom := 8;
    end; {With}
  end else begin
    if BorderStyle = bsSingle then
      Margin := SingleBorderMargin
    else
     Margin := 0;
    Arrow[1] := Point(2, Height - (8 + Margin));
    Arrow[2] := Point(8, Height - (2 + Margin));
    Arrow[3] := Point(14, Height - (8 + Margin));
    with FBottomArrowRect do begin
      Top := Height - (8 + Margin);
      Left := 2;
      Bottom := Height - (2 + Margin);
      Right := 14;
    end; {With}
  end; {If}
  C.Polygon(Arrow);
end; {DrawArrow}

procedure TCustomTimePanel.DrawConnect(C: TCanvas);
var
  Hours, TopHours: TTPDoubleHours;
  Mins: TTPMins;
  LineTop, LineBottom, LineLeft: Integer;
begin
  CalcRange;
  TopHours := (FTopHour + ((FTopDay - 1) * 24));
  with C do begin
    // Set up pen
    Pen.Style := psSolid;
    Pen.Width := FConnect.Width;
    LineLeft := 0; // Stops compiler whinging about "may not be assigned"
    case Alignment of
      taLeftJustify:
        LineLeft := TextWidth('88:88') + BoxMargin + FConnect.Width div 2;
      taCenter:
        LineLeft := (TextWidth('88:88') + Width) div 2;
      taRightJustify:
        LineLeft := ((Width - BoxMargin) - FConnect.Width div 2) -
          FConnect.Width mod 2;
    end;
    // Draw from Start to Pause
    Hours := Start.Hours;
    Mins := Start.Mins;
    LineTop := ArrowHeight + ((Hours - TopHours) * FTextHeight) -
      (Trunc(FTextHeight * (FTopMin / 60.0))) +
      (Trunc(FTextHeight * (Mins / 60.0))) +
      (TextHeight('88:88') div 2);
    Hours := Start.Hours + Pause.Hours;
    Mins := Start.Mins + Pause.Mins;
    LineBottom := ArrowHeight + ((Hours - TopHours) * FTextHeight) -
      (Trunc(FTextHeight * (FTopMin / 60.0))) +
      (Trunc(FTextHeight * (Mins / 60.0))) +
      (TextHeight('88:88') div 2);
    Pen.Color := FConnect.TopColor;
    MoveTo(LineLeft, LineTop);
    LineTo(LineLeft, LineBottom);
    // Draw from Pause to Finish
    Hours := Start.Hours + Pause.Hours;
    Mins := Start.Mins + Pause.Mins;
    LineTop := ArrowHeight + ((Hours - TopHours) * FTextHeight) -
      (Trunc(FTextHeight * (FTopMin / 60.0))) +
      (Trunc(FTextHeight * (Mins / 60.0))) +
      (TextHeight('88:88') div 2);
    Hours := Finish.Hours + ((Finish.Day - 1) * 24);
    Mins := Finish.Mins;
    LineBottom := ArrowHeight + ((Hours - TopHours) * FTextHeight) -
      (Trunc(FTextHeight * (FTopMin / 60.0))) +
      (Trunc(FTextHeight * (Mins / 60.0))) +
      (TextHeight('88:88') div 2);
    Pen.Color := FConnect.BottomColor;
    MoveTo(LineLeft, LineTop);
    LineTo(LineLeft, LineBottom);
    // Reset pen
    Pen.Width := 1;
  end; {With}
end; {DrawConnect}

function TCustomTimePanel.DrawTimeBox(C: TCanvas; Day: TTPDays; Hours: TTPHours;
  Mins: TTPMins; AColor: TColor; const BoxText: String): TRect;
const
  VMargin = 2;
  HMargin = 8;
var
  BoxRect: TRect;
  RHours, RTopHours: TTPDoubleHours;
begin
  CalcRange;
  RTopHours := (FTopHour + ((FTopDay - 1) * 24));
  RHours := (Hours + ((Day - 1) * 24));
  with C do begin
    case Alignment of
      taLeftJustify:
      begin
        BoxRect.Left := TextWidth('88:88') + BoxMargin;
        BoxRect.Right := BoxRect.Left + TextWidth(BoxText) + (2 * HMargin);
      end;
      taCenter:
      begin
        BoxRect.Left := ((TextWidth('88:88') + Width - TextWidth(BoxText)) -
          (2 * HMargin)) div 2;
        BoxRect.Right := BoxRect.Left + TextWidth(BoxText) + (2 * HMargin);
      end;
      taRightJustify:
      begin
        BoxRect.Right := Width - BoxMargin;
        BoxRect.Left := (BoxRect.Right - TextWidth(BoxText)) - (2 * HMargin);
      end;
    end;
    BoxRect.Top := 12 + ((RHours - RTopHours) * FTextHeight) -
      (Trunc(FTextHeight * (FTopMin / 60.0))) + (Trunc(FTextHeight *
      (Mins / 60.0))) - VMargin;
    BoxRect.Bottom := BoxRect.Top + TextHeight(BoxText) + VMargin;
    if Enabled then
      Brush.Color := AColor
    else
      Brush.Color := clBtnFace;
    Brush.Style := bsSolid;
    Pen.Style := psSolid;
    if Enabled then begin
      Pen.Color := clBlack;
      with BoxRect do Rectangle(Left, Top, Right, Bottom);
      Brush.Style := bsClear;
    end else begin
      Pen.Color := clBtnHighLight;
      with BoxRect do Rectangle(Left + 1, Top + 1, Right + 1, Bottom + 1);
      Brush.Style := bsClear;
      Pen.Color := clBtnShadow;
      with BoxRect do Rectangle(Left, Top, Right, Bottom);
    end;
    Font.Color := clBlack;
    TextOut(BoxRect.Left + HMargin, BoxRect.Top + (VMargin div 2), BoxText);
  end; {With}
  Result := BoxRect;
end; {DrawTimeBox}

procedure TCustomTimePanel.DrawTimeBoxes(C: TCanvas);
var
  StMins, FinMins, PMins, DMins: Integer;
  DurDay, PaDay: TTPDays;
  DurHours, PaHours: TTPHours;
  DurMins, PaMins: TTPMins;
begin
  if FSimplePanel then begin
    // Simple time box
    with Simple do
      FRect := DrawTimeBox(C, 1, Hours, Mins, Color,
        FormatDateTime(DisplayFormat, Time));
    with Simple.Rect do begin
      if (Bottom > 0) and (Top < Height) then
        Simple.FShown := True
      else
        Simple.FShown := False;
      FSimpleControl.Top := Top;
      FSimpleControl.Left := Left;
      FSimpleControl.Height := (Bottom - Top);
      FSimpleControl.Width := (Right - Left);
    end; {With}
  end else begin
    // Duration time box
    if Duration.Visible then begin
      with Start do StMins := (Hours * 60) + Mins;
      with Finish do FinMins := ((Day - 1) * (24 * 60)) + (Hours * 60) + Mins;
      with Pause do PMins := StMins + (Hours * 60) + Mins;
      DMins := PMins + ((FinMins - PMins) div 2);
      DurDay := (DMins div (24 * 60)) + 1;
      DurHours := (DMins div 60) mod 24;
      DurMins := DMins mod 60;
      with Duration do
        FRect := DrawTimeBox(C, DurDay, DurHours, DurMins, Color,
          Format(DisplayFormat, [Hours, Mins]));
      with Duration.Rect do begin
        if (Bottom > 0) and (Top < Height) then
          Duration.FShown := True
        else
          Duration.FShown := False;
        FDurationControl.Top := Top;
        FDurationControl.Left := Left;
        FDurationControl.Height := (Bottom - Top);
        FDurationControl.Width := (Right - Left);
      end; {With}
    end else
      Duration.FShown := False;
    // Pause time box
    if Pause.Visible then begin
      with Start do StMins := (Hours * 60) + Mins;
      with Pause do PMins := StMins + (Hours * 60) + Mins;
      PaDay := (PMins div (24 * 60)) + 1;
      PaHours := (PMins div 60) mod 24;
      PaMins := PMins mod 60;
      with Pause do
        FRect := DrawTimeBox(C, PaDay, PaHours, PaMins, Color,
          Format(DisplayFormat, [Hours, Mins]));
      with Pause.Rect do begin
        if (Bottom > 0) and (Top < Height) then
          Pause.FShown := True
        else
          Pause.FShown := False;
        FPauseControl.Top := Top;
        FPauseControl.Left := Left;
        FPauseControl.Height := (Bottom - Top);
        FPauseControl.Width := (Right - Left);
      end; {With}
    end else
      Pause.FShown := False;
    // Start time box
    with Start do
      FRect := DrawTimeBox(C, 1, Hours, Mins, Color,
        FormatDateTime(DisplayFormat, Time));
    with Start.Rect do begin
      if (Bottom > 0) and (Top < Height) then
        Start.FShown := True
      else
        Start.FShown := False;
      FStartControl.Top := Top;
      FStartControl.Left := Left;
      FStartControl.Height := (Bottom - Top);
      FStartControl.Width := (Right - Left);
    end; {With}
    // Finish time box
    with Finish do
      FRect := DrawTimeBox(C, Day, Hours, Mins, Color,
        FormatDateTime(DisplayFormat, Time));
    with Finish.Rect do begin
      if (Bottom > 0) and (Top < Height) then
        Finish.FShown := True
      else
        Finish.FShown := False;
      FFinishControl.Top := Top;
      FFinishControl.Left := Left;
      FFinishControl.Height := (Bottom - Top);
      FFinishControl.Width := (Right - Left);
    end; {With}
  end; {If}
end; {DrawTimeBoxes}

procedure TCustomTimePanel.DrawTimeNumbers(C: TCanvas);
var
  LineRect, TextClipRect: TRect;
  TextTop: Integer;
  HourNum, RTopHours, RBottomHours: TTPDoubleHours;
begin
  C.Font := Font;
  with C do begin
    Brush.Style := bsClear;
    with TextClipRect do begin
      Left := 0;
      Top := ArrowHeight;
      Right := TextWidth('88:88') + (2 * TimeNumMargin);
      if BorderStyle = bsSingle then
        Bottom := Height - ArrowHeight - SingleBorderMargin
      else
        Bottom := Height - ArrowHeight;
    end; {With}
    CalcRange;
    RTopHours := (FTopHour + ((FTopDay - 1) * 24));
    RBottomHours := (FBottomHour + ((FBottomDay - 1) * 24));
    for HourNum := RTopHours to RBottomHours do begin
      TextTop := TextClipRect.Top + ((HourNum - RTopHours) * FTextHeight) -
        (Trunc(FTextHeight * (FTopMin / 60.0)));
      if Enabled then begin
        if ((HourNum mod 24) >= FWorkStart) and
          ((HourNum mod 24) <= FWorkFinish) then
          Font.Color := FWorkTimeColor
        else
          Font.Color := FTimeColor;
        TextRect(TextClipRect, TimeNumMargin, TextTop, Format('%.2d:00',
          [HourNum mod 24]));
      end else begin
        Font.Color := clBtnHighlight;
        TextRect(TextClipRect, TimeNumMargin + 1, TextTop + 1, Format('%.2d:00',
          [HourNum mod 24]));
        Font.Color := clBtnShadow;
        TextRect(TextClipRect, TimeNumMargin, TextTop, Format('%.2d:00',
          [HourNum mod 24]));
      end;
    end; {For}
  end; {With}
  if FShowMargin then begin
    LineRect := GetClientRect;
    LineRect.Left := C.TextWidth('88:88') + (2 * TimeNumMargin);
    LineRect.Right := LineRect.Left + 2;
    Frame3D(C, LineRect, clBtnShadow, clBtnHighlight, 1);
  end;
end; {DrawTimeNumbers}

procedure TCustomTimePanel.CalcRange;
var
  Metric: TTextMetric;
  Margin, CalcBottom: Integer;
begin
  Margin := ArrowHeight * 2;
  if BorderStyle = bsSingle then
    Margin := Margin + SingleBorderMargin;
  GetTextMetrics(Canvas.Handle, Metric);
  FTextHeight := Metric.tmHeight * 2;
  CalcBottom := (FTopHour + ((FTopDay - 1) * 24)) +
    ((Height - Margin) div FTextHeight) + 1;
  if CalcBottom > 47 then begin
    FBottomDay := FShowDays;
    FBottomHour := 23;
  end else begin
    if CalcBottom > 23 then begin
      FBottomDay := FShowDays;
      case FShowDays of
        1: FBottomHour := 23;
        2: FBottomHour := CalcBottom;
      end; {Case}
    end else begin
      FBottomDay := 1;
      FBottomHour := CalcBottom;
    end; {If}
  end; {If}
end; {CalcRange}

procedure TCustomTimePanel.DrawFrame(C: TCanvas);
var
  TopColor, BottomColor: TColor;

  procedure AdjustColors(Bevel: TPanelBevel);
  begin
    TopColor := clBtnHighlight;
    if Bevel = bvLowered then TopColor := clBtnShadow;
    BottomColor := clBtnShadow;
    if Bevel = bvLowered then BottomColor := clBtnHighlight;
  end; {AdjustColors}

var
  Rect, MarginRect: TRect;
begin
  Rect := GetClientRect;
  if BevelOuter <> bvNone then begin
    AdjustColors(BevelOuter);
    Frame3D(C, Rect, TopColor, BottomColor, BevelWidth);
  end; {If}
  Frame3D(C, Rect, Color, Color, BorderWidth);
  if BevelInner <> bvNone then begin
    AdjustColors(BevelInner);
    Frame3D(C, Rect, TopColor, BottomColor, BevelWidth);
  end; {If}
  with C do begin
    Brush.Color := Color;
    FillRect(Rect);
  end; {With}
  C.Font := Font;
  with MarginRect do begin
    Left := Rect.Left;
    Top := Rect.Top;
    Bottom := Rect.Bottom;
    Right := C.TextWidth('88:88') + (2 * TimeNumMargin);
  end;
  with C do begin
    Brush.Color := FMarginColor;
    FillRect(MarginRect);
  end;
end; {DrawFrame}

procedure TCustomTimePanel.Paint;
var
  B: TBitmap;
  FRect: TRect;
begin
  B := TBitmap.Create;
  try
    B.Width := Width;
    B.Height := Height;
    Canvas.Font := Font;
    B.Canvas.Font := Font;
    DrawFrame(B.Canvas);
    DrawTimeNumbers(B.Canvas);
    if (not FSimplePanel) and FConnect.Visible then
      DrawConnect(B.Canvas);
    DrawTimeBoxes(B.Canvas);
    if FScrolling = sdNotScrolling then begin
      DrawArrow(B.Canvas, adUp, clBlack);
      DrawArrow(B.Canvas, adDown, clBlack);
    end else begin
      if FScrolling = sdUp then begin
        DrawArrow(B.Canvas, adUp, FSelColor);
        DrawArrow(B.Canvas, adDown, clBlack);
      end else begin
        DrawArrow(B.Canvas, adUp, clBlack);
        DrawArrow(B.Canvas, adDown, FSelColor);
      end; {If}
    end; {If}
    Canvas.Draw(0, 0, B);
    if Focused and FShowFocus then begin
      FRect := ClientRect;
      InflateRect(FRect, -4, -4);
      Canvas.Pen.Style := psSolid;
      Canvas.Pen.Color := clWindowFrame;
      Canvas.Brush.Style := bsClear;
      Canvas.Brush.Color := clBtnFace;
      DrawFocusRect(Canvas.Handle, FRect);
    end;
  finally
    B.Free;
  end; {Try..Finally}
  if FScrolling <> sdNotScrolling then begin
    if FScrolling = sdUp then begin
      Scroll(sdUp, 0, Trunc(FScrollAmount));
      Invalidate;
      Sleep(5);
    end else begin
      Scroll(sdDown, 0, Trunc(FScrollAmount));
      Invalidate;
      Sleep(5);
    end; {If}
    FScrollAmount := FScrollAmount + 0.5;
    if FScrollAmount > 10.0 then // 15/10/99 DJ
      FScrollAmount := 10.0;  // 15/10/99 DJ
  end; {If}
end; {Paint}

{ TCustomTimePanel ---------------------------------------- Message Handlers --}

procedure TCustomTimePanel.CMEnabledChanged(var Message: TMessage);
begin
  Invalidate;

  inherited;
end; {CMEnabledChanged}

procedure TCustomTimePanel.WMKillFocus(var Message: TMessage);
begin
  if FShowFocus then Invalidate;

  inherited;
end; {WMKillFocus}

procedure TCustomTimePanel.WMSetFocus(var Message: TMessage);
begin
  if FShowFocus then Invalidate;

  inherited;
end; {WMSetFocus}

{ TCustomTimePanel --------------------------------------- Colour Properties --}

procedure TCustomTimePanel.SetTimeColor(Value: TColor);
begin
  if Value <> FTimeColor then begin
    FTimeColor := Value;
    Invalidate;
  end;
end; {SetTimeColor}

procedure TCustomTimePanel.SetWorkTimeColor(Value: TColor);
begin
  if Value <> FWorkTimeColor then begin
    FWorkTimeColor := Value;
    Invalidate;
  end;
end; {SetWorkTimeColor}

procedure TCustomTimePanel.SetSelColor(Value: TColor);
begin
  if Value <> FSelColor then begin
    FSelColor := Value;
    Invalidate;
  end;
end; {SetSelColor}

procedure TCustomTimePanel.SetMarginColor(Value: TColor);
begin
  if Value <> FMarginColor then begin
    FMarginColor := Value;
    Invalidate;
  end;
end; {SetMarginColor}

{ TCustomTimePanel --------------------------------------- Scroll Properties --}

procedure TCustomTimePanel.SetTopDay(Value: TTPDays);
begin
  if Value <> FTopDay then begin
    if Value <= FShowDays then begin
      FTopDay := Value;
      Invalidate;
    end; {If}
  end; {If}
end; {SetTopDay}

procedure TCustomTimePanel.SetTopHour(Value: TTPHours);
begin
  if Value <> FTopHour then begin
    FTopHour := Value;
    Invalidate;
  end;
end; {SetTopHour}

procedure TCustomTimePanel.SetTopMin(Value: TTPMins);
begin
  if Value <> FTopMin then begin
    FTopMin := Value;
    Invalidate;
  end;
end; {SetTopMin}

{ TCustomTimePanel ----------------------------------------- Show Properties --}

procedure TCustomTimePanel.SetShowDays(Value: TTPDays);
begin
  if (not FSimplePanel) and (Value <> FShowDays) then begin
    FShowDays := Value;
    if (Finish.Day = 2) and (FShowDays = 1) then begin
      Finish.Day := 1;
      Invalidate;
      DoChange;
    end; {If}
  end; {If}
end; {SetShowDays}

procedure TCustomTimePanel.SetSimplePanel(Value: Boolean);
begin
  if Value <> FSimplePanel then begin
    FSimplePanel := Value;
    if FSimplePanel then FShowDays := 1;
    Invalidate;
  end;
end; {SetSimplePanel}

procedure TCustomTimePanel.SetShowMargin(Value: Boolean);
begin
  if Value <> FShowMargin then begin
    FShowMargin := Value;
    Invalidate;
  end;
end; {SetShowMargin}

procedure TCustomTimePanel.SetShowFocus(Value: Boolean);
begin
  if Value <> FShowFocus then begin
    FShowFocus := Value;
    Invalidate;
  end;
end; {SetShowFocus}

{ TCustomTimePanel -------------------------------- Miscellaneous Properties --}

procedure TCustomTimePanel.SetWorkStart(Value: TTPHours);
begin
  if Value <> FWorkStart then begin
    if Value <= FWorkFinish then
      FWorkStart := Value
    else
      FWorkStart := FWorkFinish;
    Invalidate;
    DoChange;
  end;
end; {SetWorkStart}

procedure TCustomTimePanel.SetWorkFinish(Value: TTPHours);
begin
  if Value <> FWorkFinish then
  begin
    if Value >= FWorkStart then
      FWorkFinish := Value
    else
      FWorkFinish := FWorkStart;
    Invalidate;
    DoChange;
  end;
end; {SetWorkFinish}

{ TCustomTimePanel -------------------------------------- Scrolling Handlers --}

function TCustomTimePanel.IsInRect(ARect: TRect; X, Y: Integer): Boolean;
begin
  with ARect do
    Result := (X >= Left) and (X <= Right) and (Y >= Top) and (Y <= Bottom);
end; {IsInRect}

procedure TCustomTimePanel.Scroll(Dir: TTPScrollDirection; Hours: TTPHours;
  Mins: TTPMins);
var
  TopMins, ScrollMins: Integer;
begin
  TopMins := ((FTopDay - 1) * (24 * 60)) + (FTopHour * 60) + FTopMin;
  ScrollMins := (Hours * 60) + Mins;
  if Dir = sdUp then begin
    Dec(TopMins, ScrollMins);
    if TopMins < 0 then
      TopMins := 0;
  end else begin
    Inc(TopMins, ScrollMins);
    if FShowDays = 1 then begin
      if TopMins > ((24 * 60) - 1) then
        TopMins := (24 * 60) - 1;
    end else begin
      if TopMins > ((24 * 60 * 2) - 1) then
        TopMins := (24 * 60 * 2) - 1;
    end; {If}
  end; {If}
  FTopDay := (TopMins div (24 * 60)) + 1;
  FTopHour := (TopMins div 60) mod 24;
  FTopMin := TopMins mod 60;
  Invalidate;
end; {Scroll}

procedure TCustomTimePanel.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if Shift = [ssLeft] then
    if IsInRect(FTopArrowRect, X, Y) then begin
      FScrolling := sdUp;
      FScrollAmount := 1.0;
      Invalidate;
    end else if IsInRect(FBottomArrowRect, X, Y) then begin
      FScrolling := sdDown;
      FScrollAmount := 1.0;
      Invalidate;
    end else
      FScrolling := sdNotScrolling;
end; {MouseDown}

procedure TCustomTimePanel.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if (FScrolling = sdNotScrolling) and (Shift = []) then
    case Key of
      VK_PRIOR:
      begin
        FScrolling := sdUp;
        FScrollAmount := 1.0;
        Invalidate;
        Key := 0;
      end;
      VK_NEXT:
      begin
        FScrolling := sdDown;
        FScrollAmount := 1.0;
        Invalidate;
        Key := 0;
      end;
    else
      FScrolling := sdNotScrolling;
    end;
  inherited KeyDown(Key, Shift);
end; {KeyDown}

procedure TCustomTimePanel.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  if (FScrolling <> sdNotScrolling) and (Button = mbLeft) then
    FScrolling := sdNotScrolling;
end; {MouseUp}

procedure TCustomTimePanel.KeyUp(var Key: Word; Shift: TShiftState);
begin
  if (FScrolling <> sdNotScrolling) and (Shift = []) and
    (Key in [VK_PRIOR, VK_NEXT]) then
    FScrolling := sdNotScrolling;
  inherited KeyUp(Key, Shift);
end; {KeyUp}

{ TCustomTimePanel ------------------------------------- Mouse Drag Handlers --}

procedure TCustomTimePanel.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
  // Deal with scrolling arrows
  if IsInRect(FTopArrowRect, X, Y) then
    DrawArrow(Canvas, adUp, SelColor)
  else
    DrawArrow(Canvas, adUp, clBlack);
  if IsInRect(FBottomArrowRect, X, Y) then
    DrawArrow(Canvas, adDown, SelColor)
  else
    DrawArrow(Canvas, adDown, clBlack);
end;

procedure TCustomTimePanel.StartMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
  OldTop, NewTop: Integer;
  Hours: TTPHours;
  Mins: TTPMins;
begin
  FScrolling := sdNotScrolling; // 15/10/99 DJ
  if (not FSimplePanel) and (Shift = [ssLeft]) then
  begin
    OldTop := FStartControl.Top;
    NewTop := FStartControl.Top + Y;
    if Y > FStartControl.Height then Dec(NewTop, FStartControl.Height);
    FStartControl.Top := NewTop;
    Hours := (Abs(OldTop - NewTop) div FTextHeight) mod 24;
    Mins := Trunc((Abs(OldTop - NewTop) / FTextHeight) * 12) * FTimeShift;
    if Y < 0 then
      Start.DecTime(Hours, Mins + (Start.Mins mod FTimeShift))
    else if Y > FStartControl.Height then
      Start.IncTime(Hours, Mins + (FTimeShift - (Start.Mins mod FTimeShift)));
    // 15/10/99 DJ -------------------------------------------------------------
    if (FStartControl.Top + FStartControl.Height) > Height then begin
      FScrolling := sdDown;
      Invalidate;
    end else if FStartControl.Top < 0 then begin
      FScrolling := sdUp;
      Invalidate;
    end else
      FScrolling := sdNotScrolling;
    // 15/10/99 DJ -------------------------------------------------------------
  end; {If}
end; {StartMouseMove}

procedure TCustomTimePanel.FinishMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
  OldTop, NewTop: Integer;
  Hours: TTPHours;
  Mins: TTPMins;
begin
  FScrolling := sdNotScrolling; // 15/10/99 DJ
  if (not FSimplePanel) and (Shift = [ssLeft]) then
  begin
    OldTop := FFinishControl.Top;
    NewTop := FFinishControl.Top + Y;
    if Y > FFinishControl.Height then Dec(NewTop, FFinishControl.Height);
    FFinishControl.Top := NewTop;
    Hours := (Abs(OldTop - NewTop) div FTextHeight) mod 24;
    Mins := Trunc((Abs(OldTop - NewTop) / FTextHeight) * 12) * FTimeShift;
    if Y < 0 then
      Finish.DecTime(Hours, Mins + (Finish.Mins mod FTimeShift))
    else if Y > FFinishControl.Height then
      Finish.IncTime(Hours, Mins + (FTimeShift - (Finish.Mins mod FTimeShift)));
    // 15/10/99 DJ -------------------------------------------------------------
    if (FFinishControl.Top + FFinishControl.Height) > Height then begin
      FScrolling := sdDown;
      Invalidate;
    end else if FFinishControl.Top < 0 then begin
      FScrolling := sdUp;
      Invalidate;
    end else
      FScrolling := sdNotScrolling;
    // 15/10/99 DJ -------------------------------------------------------------
  end; {If}
end; {FinishMouseMove}

procedure TCustomTimePanel.DurationMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
  OldTop, NewTop: Integer;
  Hours: TTPHours;
  Mins: TTPMins;
begin
  FScrolling := sdNotScrolling; // 15/10/99 DJ
  if (not FSimplePanel) and (Shift = [ssLeft]) and (Duration.Visible) then begin
    OldTop := FDurationControl.Top;
    NewTop := FDurationControl.Top + Y;
    if Y > FDurationControl.Height then Dec(NewTop, FDurationControl.Height);
    FDurationControl.Top := NewTop;
    Hours := (Abs(OldTop - NewTop) div FTextHeight) mod 24;
    Mins := Trunc((Abs(OldTop - NewTop) / FTextHeight) * 12) * FTimeShift;
    if Y < 0 then
      Duration.DecTime(Hours, Mins)
    else if Y > FDurationControl.Height then
      Duration.IncTime(Hours, Mins);
    // 15/10/99 DJ -------------------------------------------------------------
    if (FDurationControl.Top + FDurationControl.Height) > Height then begin
      FScrolling := sdDown;
      Invalidate;
    end else if FDurationControl.Top < 0 then begin
      FScrolling := sdUp;
      Invalidate;
    end else
      FScrolling := sdNotScrolling;
    // 15/10/99 DJ -------------------------------------------------------------
  end; {If}
end; {DurationMouseMove}

procedure TCustomTimePanel.PauseMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
  OldTop, NewTop: Integer;
  Hours: TTPHours;
  Mins: TTPMins;
begin
  FScrolling := sdNotScrolling; // 15/10/99 DJ
  if (not FSimplePanel) and (Shift = [ssLeft]) and (Pause.Visible) then begin
    OldTop := FPauseControl.Top;
    NewTop := FPauseControl.Top + Y;
    if Y > FPauseControl.Height then Dec(NewTop, FPauseControl.Height);
    FPauseControl.Top := NewTop;
    Hours := (Abs(OldTop - NewTop) div FTextHeight) mod 24;
    Mins := Trunc((Abs(OldTop - NewTop) / FTextHeight) * 12) * FTimeShift;
    if Y < 0 then
      Pause.DecTime(Hours, Mins)
    else if Y > FPauseControl.Height then
      Pause.IncTime(Hours, Mins);
    // 15/10/99 DJ -------------------------------------------------------------
    if (FPauseControl.Top + FPauseControl.Height) > Height then begin
      FScrolling := sdDown;
      Invalidate;
    end else if FPauseControl.Top < 0 then begin
      FScrolling := sdUp;
      Invalidate;
    end else
      FScrolling := sdNotScrolling;
    // 15/10/99 DJ -------------------------------------------------------------
  end; {If}
end; {PauseMouseMove}

procedure TCustomTimePanel.SimpleMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
  OldTop, NewTop: Integer;
  Hours: TTPHours;
  Mins: TTPMins;
begin
  FScrolling := sdNotScrolling; // 15/10/99 DJ
  if (FSimplePanel) and (Shift = [ssLeft]) then begin
    OldTop := FSimpleControl.Top;
    NewTop := FSimpleControl.Top + Y;
    if Y > FSimpleControl.Height then Dec(NewTop, FSimpleControl.Height);
    FSimpleControl.Top := NewTop;
    Hours := (Abs(OldTop - NewTop) div FTextHeight) mod 24;
    Mins := Trunc((Abs(OldTop - NewTop) / FTextHeight) * 12) * FTimeShift;
    if Y < 0 then
      Simple.DecTime(Hours, Mins + (Simple.Mins mod FTimeShift))
    else if Y > FSimpleControl.Height then
      Simple.IncTime(Hours, Mins + (FTimeShift - (Simple.Mins mod FTimeShift)));
    // 15/10/99 DJ -------------------------------------------------------------
    if (FSimpleControl.Top + FSimpleControl.Height) > Height then begin
      FScrolling := sdDown;
      Invalidate;
    end else if FSimpleControl.Top < 0 then begin
      FScrolling := sdUp;
      Invalidate;
    end else
      FScrolling := sdNotScrolling;
    // 15/10/99 DJ -------------------------------------------------------------
  end; {If}
end; {SimpleMouseMove}

{------------------------------------------------------------ Event Handlers --}

procedure TCustomTimePanel.DoChange;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end; {DoChange}

{ *************************************************************** TDBTimePanel }

constructor TDBTimePanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  // Create data links
  FStartData := TFieldDataLink.Create;
  with FStartData do begin
    Control := Self;
    OnDataChange := DataChange;
    OnUpdateData := UpdateData;
  end;
  FFinishData := TFieldDataLink.Create;
  with FFinishData do begin
    Control := Self;
    OnDataChange := DataChange;
    OnUpdateData := UpdateData;
  end;
  FDurationData := TFieldDataLink.Create;
  with FDurationData do begin
    Control := Self;
    OnDataChange := DataChange;
    OnUpdateData := UpdateData;
  end;
  FPauseData := TFieldDataLink.Create;
  with FPauseData do begin
    Control := Self;
    OnDataChange := DataChange;
    OnUpdateData := UpdateData;
  end;
  FSimpleData := TFieldDataLink.Create;
  with FSimpleData do begin
    Control := Self;
    OnDataChange := DataChange;
    OnUpdateData := UpdateData;
  end;
  FWriting := False;
end; {Create}

destructor TDBTimePanel.Destroy;
begin
  FStartData.Free;
  FFinishData.Free;
  FDurationData.Free;
  FPauseData.Free;
  FSimpleData.Free;
  FStartData := nil;
  FFinishData := nil;
  FDurationData := nil;
  FPauseData := nil;
  FSimpleData := nil;
  inherited Destroy;
end; {Destroy}

procedure TDBTimePanel.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (FStartData <> nil) and
    (FFinishData <> nil) and (FDurationData <> nil) and
    (FPauseData <> nil) and (FSimpleData <> nil) and
    (AComponent = DataSource) then DataSource := nil;
end; {Notification}

{ TDBTimePanel --------------------------------------- DataSource Properties --}

function TDBTimePanel.GetDataSource: TDataSource;
begin
  // Just pick a field ... start'll do!
  Result := FStartData.DataSource;
end; {GetDataSource}

procedure TDBTimePanel.SetDataSource(Value: TDataSource);
begin
  // Remember to set ALL field's datasources...
  FStartData.DataSource := Value;
  FFinishData.DataSource := Value;
  FPauseData.DataSource := Value;
  FDurationData.DataSource := Value;
  FSimpleData.DataSource := Value;
  if Value <> nil then Value.FreeNotification(Self);
end; {SetDataSource}

function TDBTimePanel.GetReadOnly: Boolean;
begin
  Result := FStartData.ReadOnly;
end; {GetReadOnly}

procedure TDBTimePanel.SetReadOnly(Value: Boolean);
begin
  FDurationData.ReadOnly := Value;
  FFinishData.ReadOnly := Value;
  FPauseData.ReadOnly := Value;
  FSimpleData.ReadOnly := Value;
  FStartData.ReadOnly := Value;
end; {SetReadOnly}

{ TDBTimePanel ---------------------------------------- DataField Properties --}

function TDBTimePanel.GetStartDataField: String;
begin
  Result := FStartData.FieldName;
end; {GetStartDataField}

procedure TDBTimePanel.SetStartDataField(Value: String);
begin
  FStartData.FieldName := Value;
end; {SetStartDataField}

function TDBTimePanel.GetDurationDataField: String;
begin
  Result := FDurationData.FieldName;
end; {GetDurationDataField}

procedure TDBTimePanel.SetDurationDataField(Value: String);
begin
  FDurationData.FieldName := Value;
end; {SetDurationDataField}

function TDBTimePanel.GetFinishDataField: String;
begin
  Result := FFinishData.FieldName;
end; {GetFinishDataField}

procedure TDBTimePanel.SetFinishDataField(Value: String);
begin
  FFinishData.FieldName := Value;
end; {SetFinishDataField}

function TDBTimePanel.GetPauseDataField: String;
begin
  Result := FPauseData.FieldName;
end; {GetPauseDataField}

procedure TDBTimePanel.SetPauseDataField(Value: String);
begin
  FPauseData.FieldName := Value;
end; {SetPauseDataField}

function TDBTimePanel.GetSimpleDataField: String;
begin
  Result := FSimpleData.FieldName;
end; {GetSimpleDataField}

procedure TDBTimePanel.SetSimpleDataField(Value: String);
begin
  FSimpleData.FieldName := Value;
end; {SetSimpleDataField}

{ TDBTimePanel -------------------------------------------- Field Properties --}

function TDBTimePanel.GetDurationField: TField;
begin
  Result := FDurationData.Field;
end; {GetDurationField}

function TDBTimePanel.GetFinishField: TField;
begin
  Result := FFinishData.Field;
end; {GetFinishField}

function TDBTimePanel.GetPauseField: TField;
begin
  Result := FPauseData.Field;
end; {GetPauseField}

function TDBTimePanel.GetSimpleField: TField;
begin
  Result := FSimpleData.Field;
end; {GetSimpleField}

function TDBTimePanel.GetStartField: TField;
begin
  Result := FStartData.Field;
end; {GetStartField}

{ TDBTimePanel -------------------------------------------- Message Handlers --}

procedure TDBTimePanel.CMExit(var Message: TCMExit);
begin
  FStartData.UpdateRecord;
  FFinishData.UpdateRecord;
  FDurationData.UpdateRecord;
  FSimpleData.UpdateRecord;
  FPauseData.UpdateRecord;
  inherited;
end; {CMExit}

procedure TDBTimePanel.CMGetDataLink(var Message: TMessage);
begin
  Message.Result := Integer(FStartData);
end; {CMGetDataLink}

{ TDBTimePanel ----------------------------------------------- Data Handlers --}

function TDBTimePanel.CanEdit(ControlKind: TTPControlKind): Boolean;
begin
  Result := False;
  case ControlKind of
    ckStart:
      if FStartData.Edit then begin
        Result := True;
        FStartData.Modified;
      end else if (Pause.Hours = 0) and (Pause.Mins = 0) then begin
        if FFinishData.Edit and FDurationData.Edit then begin
          Result := True;
          FFinishData.Modified;
          FDurationData.Modified;
        end;
      end else begin
        if FFinishData.Edit and FDurationData.Edit and FPauseData.Edit then begin
          Result := True;
          FFinishData.Modified;
          FDurationData.Modified;
          FPauseData.Modified;
        end;
      end;
    ckFinish:
      if FFinishData.Edit then begin
        Result := True;
        FFinishData.Modified;
      end else if (Pause.Hours = 0) and (Pause.Mins = 0) then begin
        if FStartData.Edit and FDurationData.Edit then begin
          Result := True;
          FStartData.Modified;
          FDurationData.Modified;
        end;
      end else begin
        if FStartData.Edit and FDurationData.Edit and FPauseData.Edit then begin
          Result := True;
          FStartData.Modified;
          FDurationData.Modified;
          FPauseData.Modified;
        end;
      end;
    ckDuration:
      if FDurationData.Edit then begin
        Result := True;
        FDurationData.Modified;
      end else if (Pause.Hours = 0) and (Pause.Mins = 0) then begin
        if FFinishData.Edit and FStartData.Edit then begin
          Result := True;
          FFinishData.Modified;
          FStartData.Modified;
        end;
      end else begin
        if FFinishData.Edit and FStartData.Edit and FPauseData.Edit then begin
          Result := True;
          FFinishData.Modified;
          FStartData.Modified;
          FPauseData.Modified;
        end;
      end;
    ckPause:
      if FPauseData.Edit then begin
        Result := True;
        FPauseData.Modified;
      end else begin
        if FFinishData.Edit and FDurationData.Edit and FStartData.Edit then begin
          Result := True;
          FFinishData.Modified;
          FDurationData.Modified;
          FStartData.Modified;
        end;
      end;
    ckSimple:
      if FSimpleData.Edit then begin
        Result := True;
        FSimpleData.Modified;
      end;
  else
    Result := False;
  end; {Case}
end; {CanEdit}

procedure TDBTimePanel.DataChange(Sender: TObject);

  function GetMins(DataLink: TFieldDataLink; var Mins: Integer): Boolean;
  var
    H, M, S, MS: Word;
  begin
    Result := True;
    if DataLink.Field <> nil then
      if not DataLink.Field.IsNull then
        Result := False;
    if not Result then begin
      case DataLink.Field.DataType of
        ftDateTime, ftTime:
          DecodeTime(DataLink.Field.AsDateTime, H, M, S, MS);
        ftFloat:
          DecodeTime(DataLink.Field.AsFloat, H, M, S, MS);
        ftString:
          try
            DecodeTime(StrToTime(DataLink.Field.AsString), H, M, S, MS);
          except
            Result := True;
          end;
      else
        Result := True;
      end; {Case}
      if not Result then begin
        Mins := (H * 60) + M;
        if Mins < 0 then Mins := 0;
      end;
    end; {If}
  end; {GetMins}

  procedure SetMins(Control: TCustomTimeBox; Mins: Integer);
  begin
    Control.FHours := Mins div 60;
    Control.FMins := Mins mod 60;
  end; {SetMins}

var
  StMins, FinMins, PaMins, DurMins, SiMins: Integer;
  StNull, FinNull, PaNull, DurNull, SiNull: Boolean;
  NumFields: Integer;
begin
  // See note about FWriting in UpdateData
  if not FWriting then
    if FSimplePanel then begin
      // Simple panel stuff
      SiNull := GetMins(FSimpleData, SiMins);
      if not SiNull then begin
        FSimple.FHours := SiMins div 60;
        FSimple.FMins := SiMins mod 60;
        Invalidate;
        DoChange;
      end else
        SiMins := 0;
      TopMin := 0;
      if (SiMins div 60) = 0 then
        TopHour := 0 else
        TopHour := (SiMins div 60) - 1;
    end else begin
      // Get values from fields (if they exist)
      StNull := GetMins(FStartData, StMins);
      FinNull := GetMins(FFinishData, FinMins);
      PaNull := GetMins(FPauseData, PaMins);
      DurNull := GetMins(FDurationData, DurMins);
      // Work out how many values we've got to play with
      NumFields := Ord(not StNull) + Ord(not FinNull) + Ord(not PaNull) +
        Ord(not DurNull);
      // Figure out values of missing fields
      case NumFields of
        0:
        begin
          // If we have no fields available (all missing or null), then all
          // times are zero (start, finish, duration etc.)
          StMins := 0;
          FinMins := 0;
          DurMins := 0;
          PaMins := 0;
        end;
        1:
        begin
          // If we have one value to deal with, make up the rest (usually zero,
          // unless it's Duration and Finish which may take their values from
          // the field that we _do_ have)
          if StNull then StMins := 0 else FinMins := StMins;
          if PaNull then PaMins := 0 else FinMins := PaMins;
          if DurNull and FinNull then DurMins := 0;
          if not DurNull then FinMins := DurMins;
          if not FinNull then DurMins := FinMins;
        end;
        2:
        begin
          // If we have two values to deal with, make up the rest (usually zero)
          if PaNull then PaMins := 0;
          if DurNull and (StNull or FinNull) then DurMins := 0;

          if StNull and DurNull then begin
            if (FinMins - PaMins) < 0 then Inc(FinMins, 24 * 60);
            StMins := FinMins - PaMins;
          end
          else if PaNull and DurNull then begin
            if (FinMins < StMins) then Inc(FinMins, 24 * 60);
            DurMins := FinMins - StMins;
          end
          else if StNull and PaNull then begin
            if (FinMins - DurMins) < 0 then Inc(FinMins, 24 * 60);
            StMins := FinMins - DurMins;
          end
          else if FinNull and PaNull then FinMins := StMins + DurMins
          else if FinNull and DurNull then FinMins := StMins + PaMins;
        end;
        3:
        begin
          // If we have three values to play with, the final one can easily be
          // derived from the others
          if StNull then begin
            if ((FinMins - DurMins) - PaMins) < 0 then Inc(FinMins, 24 * 60);
            StMins := ((FinMins - DurMins) - PaMins);
          end
          else if DurNull then begin
            if FinMins < StMins then Inc(FinMins, 24 * 60);
            DurMins := (FinMins - StMins) - PaMins;
          end
          else if PaNull then begin
            if FinMins < StMins then Inc(FinMins, 24 * 60);
            PaMins := (FinMins - StMins) - DurMins;
          end
          else if FinNull then FinMins := StMins + PaMins + DurMins;
        end;
        4:
        begin
          // If we have all the values to play with, still calculate one of the
          // values to ensure that everything adds up correctly
          FinMins := StMins + DurMins + PaMins;
        end;
      end; {Case}
      // If we have calculated values, it is possible that FinMins goes over the
      // midnight line when ShowDays is set to 1 - so correct for it...
      if (ShowDays = 1) and (FinMins > ((24 * 60) - 1)) then begin
        FinMins := (24 * 60) - 1;
        // Try decreasing pause before decreasing duration
        PaMins := (FinMins - StMins) - DurMins;
        // If that fails, set pause to zero and decrease duration
        if PaMins < 0 then begin
          PaMins := 0;
          DurMins := (FinMins - StMins);
        end;
      end; {If}
      SetMins(FStart, StMins);
      SetMins(FDuration, DurMins);
      SetMins(FPause, PaMins);
      SetMins(FFinish, FinMins);
      if FinMins > ((24 * 60) - 1) then
        FFinish.FDay := 2 else
        FFinish.FDay := 1;
      Invalidate;
      DoChange;
      TopMin := 0;
      if (StMins div 60) = 0 then
        TopHour := 0 else
        TopHour := (StMins div 60) - 1;
    end; {If}
end; {DataChange}

procedure TDBTimePanel.UpdateData(Sender: TObject);

  procedure WriteData(DataLink: TFieldDataLink; T: TDateTime);
  begin
    if DataLink <> nil then
      if DataLink.Field <> nil then
        case DataLink.Field.DataType of
          ftDateTime, ftTime: DataLink.Field.AsDateTime := T;
          ftFloat: DataLink.Field.AsFloat := T;
          ftString: DataLink.Field.AsString := TimeToStr(T);
        end;
  end; {WriteData}

begin
  // Set FWriting to prevent DataChange from reading back data and calculating
  // fields while we update the table
  FWriting := True;
  WriteData(FStartData, FStart.GetTime);
  WriteData(FFinishData, FFinish.GetTime);
  WriteData(FDurationData, FDuration.GetTime);
  WriteData(FPauseData, FPause.GetTime);
  WriteData(FSimpleData, FSimple.GetTime);
  FWriting := False;
end; {UpdateData}

{ ********************************************************* TDataFieldProperty }

{---
  Note : This stuff was copied (almost) verbatim from the D2 DBREG.PAS, in
  order that I can make the DataField properties in the TDBTimePanel class into
  TDataFieldPropertys in the Object Inspector instead of TStringProperty.
  Unfortunately, TDataFieldProperty is declared privately in DBREG.PAS, hence
  the reason I need to copy it, and it's ancestor TDBStringProperty
---}

{ TDBStringProperty }

{$HINTS OFF} // This is to stop the D3 (and above) compiler whinging about
             // GetValueList not being used or exported
type
  TDBStringProperty = class(TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValueList(List: TStrings); virtual; abstract;
    procedure GetValues(Proc: TGetStrProc); override;
  end;

function TDBStringProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paValueList, paSortList, paMultiSelect];
end;

procedure TDBStringProperty.GetValues(Proc: TGetStrProc);
var
  I: Integer;
  Values: TStringList;
begin
  Values := TStringList.Create;
  try
    GetValueList(Values);
    for I := 0 to Values.Count - 1 do Proc(Values[I]);
  finally
    Values.Free;
  end;
end;

{ TDataFieldProperty }

type
  TDataFieldProperty = class(TDBStringProperty)
  public
    function GetDataSourcePropName: string; virtual;
    procedure GetValueList(List: TStrings); override;
  end;

function TDataFieldProperty.GetDataSourcePropName: string;
begin
  Result := 'DataSource';
end;

procedure TDataFieldProperty.GetValueList(List: TStrings);
var
  Instance: TComponent;
  PropInfo: PPropInfo;
  DataSource: TDataSource;
begin
  Instance := (GetComponent(0) as TComponent);
  PropInfo := TypInfo.GetPropInfo(Instance.ClassInfo, GetDataSourcePropName);
  if (PropInfo <> nil) and (PropInfo^.PropType^.Kind = tkClass) then
  begin
    DataSource := TObject(GetOrdProp(Instance, PropInfo)) as TDataSource;
    if (DataSource <> nil) and (DataSource.DataSet <> nil) then
      DataSource.DataSet.GetFieldNames(List);
  end;
end;

{ ***************************************************** Component Registration }

procedure Register;
begin
  RegisterComponents('Time', [TTimePanel, TDBTimePanel]);

  RegisterPropertyEditor(TypeInfo(TTimeBox), TCustomTimePanel, '',
    TClassProperty);
  RegisterPropertyEditor(TypeInfo(TTimeBoxConnect), TCustomTimePanel, '',
    TClassProperty);

  RegisterPropertyEditor(TypeInfo(String), TDBTimePanel, 'DurationDataField',
    TDataFieldProperty);
  RegisterPropertyEditor(TypeInfo(String), TDBTimePanel, 'FinishDataField',
    TDataFieldProperty);
  RegisterPropertyEditor(TypeInfo(String), TDBTimePanel, 'PauseDataField',
    TDataFieldProperty);
  RegisterPropertyEditor(TypeInfo(String), TDBTimePanel, 'SimpleDataField',
    TDataFieldProperty);
  RegisterPropertyEditor(TypeInfo(String), TDBTimePanel, 'StartDataField',
    TDataFieldProperty);
end; {Register}

{$IFDEF D2}{$UNDEF D2}{$ENDIF}
{$IFDEF D3}{$UNDEF D3}{$ENDIF}
{$IFDEF D4}{$UNDEF D4}{$ENDIF}

end.
